"""
Entry point for standalone brewblox_service.

This will create the app, and run a simulator.
"""

from brewblox_service import events, service, simulator


def main():
    app = service.create_app(default_name='simulator')

    # Add implementation-specific functionality
    # In this case: simulator
    simulator.setup(app)

    # Event handling is optional
    # It should be enabled explicitly by service implementations
    events.setup(app)

    # Add all default endpoints
    service.furnish(app)

    # service.run() will start serving clients async
    service.run(app)


if __name__ == '__main__':
    main()



"""
Default service implementation: a simulator
"""

import logging
import random
from pprint import pformat
from typing import Type

import dpath
from aiohttp import web

LOGGER = logging.getLogger(__name__)
routes = web.RouteTableDef()


def setup(app: Type[web.Application]):
    app.router.add_routes(routes)
    app['simulator'] = SimulatorService()


class SimulatorService():
    """Responsible for making functionality available to the app.

    Keeps an internal 'config' object to mimic block configuration.
    The config makes some assumptions:
    * There are unique 'identifier' keys.
    * Volatile state objects are contained in a 'state' dict

    Example config:
    [
        {
            'identifier': 'unique_name',
            'state': {
                'value1': 1.2,
                'value2': True,
                'value3', 'stuff
            }
        }
    ]
    """

    def __init__(self):
        self._config = {}

    @property
    def config(self):
        return self._config

    @config.setter
    def config(self, val: dict):
        assert val
        LOGGER.info(f'updating simulator config to \n{pformat(val)}')
        self._config = val

    def randomize(self, config: dict):
        for k, v in dpath.util.search(config, '**/state/*', yielded=True):
            new_val = 'value'
            if isinstance(v, bool):
                new_val = random.choice([True, False])
            elif isinstance(v, (int, float)):  # pragma: no cover (incorrectly flagged as not covered)
                new_val = random.random() / random.random()
            dpath.util.new(config, k, new_val)

    def filter(self, config: dict, id: str):
        for k, v in dpath.util.search(config, '**/identifier', yielded=True):
            if v == id:
                # return parent element of correct identifier
                return dpath.util.get(config, k.split('/')[:-1])
        # should not throw if identifier is absent
        return dict()


@routes.get('/config')
async def get_config(request):
    """
    ---
    tags:
    - Simulator
    summary: Get current config.
    description: Get config currently used by simulator.
    operationId: simulator.config
    produces:
    - application/json
    """
    sim = request.app['simulator']
    return web.json_response(sim.config)


@routes.post('/config')
async def post_config(request):
    """
    ---
    tags:
    - Simulator
    summary: Update config.
    description: Set new configuration to be returned by simulator.
    operationId: simulator.config
    produces:
    - application/json
    parameters:
    -
        in: body
        name: body
        description: New configuration
        required: true
        schema:
            type: array
            items:
                properties:
                    identifier:
                        type: string
                    name:
                        type: string
                    type_id:
                        type: integer
                        format: int64
                    state:
                        type: object
                    settings:
                        type: object
    """
    sim = request.app['simulator']
    sim.config = await request.json()
    return web.json_response(dict(success=True))


@routes.get('/values')
async def get_values(request):
    """
    ---
    tags:
    - Simulator
    summary: Get all config values.
    description: Get randomized state for all values currently in config.
    operationId: simulator.values
    produces:
    - application/json
    """
    sim = request.app['simulator']
    sim.randomize(sim.config)
    return web.json_response(sim.config)


@routes.get('/values/{id}')
async def get_specific_value(request):
    """
    ---
    tags:
    - Simulator
    summary: Get all config values.
    description: Get randomized state for all values matching a specific identifier.
    operationId: simulator.values.id
    produces:
    - application/json
    parameters:
    -
        name: id
        in: path
        required: true
        description: the unique identifier of the desired value
        schema:
            type: string
    """
    id = request.match_info['id']
    sim = request.app['simulator']
    output = sim.filter(sim.config, id)
    sim.randomize(output)
    return web.json_response(output)



"""
Offers event publishing and subscription for service implementations.

Example use:

    events.setup(app)

    async def on_message(queue, key, message):
        logging.info(f'Message from {queue}: {key} = {message} ({type(message)})')

    listener = events.get_listener(app)
    listener.subscribe('brewblox', 'controller', on_message=on_message)
    listener.subscribe('brewblox', 'controller.*', on_message=on_message)
    listener.subscribe('brewblox', 'controller.#', on_message=on_message)
"""

import asyncio
import json
import logging
from concurrent.futures import CancelledError
from typing import Callable, Type, Union

import aio_pika
from aio_pika import ExchangeType
from aio_pika.queue import ExchangeType_
from aiohttp import web

LOGGER = logging.getLogger(__name__)
routes = web.RouteTableDef()

EVENT_CALLBACK_ = Callable[['EventSubscription', str, Union[dict, str]], None]

LISTENER_KEY = 'events.listener'
PUBLISHER_KEY = 'events.publisher'
EVENTBUS_HOST = 'eventbus'


def setup(app: Type[web.Application]):
    app[LISTENER_KEY] = EventListener(app)
    app[PUBLISHER_KEY] = EventPublisher(app)
    app.router.add_routes(routes)


def get_listener(app: Type[web.Application]) -> 'EventListener':
    return app[LISTENER_KEY]


def get_publisher(app: Type[web.Application]) -> 'EventPublisher':
    return app[PUBLISHER_KEY]


def _default_on_message(queue: Type[aio_pika.Queue], key: str, message: Union[dict, str]):
    LOGGER.info(f'Unhandled event: queue={queue}, key={key} message={message}')


class EventSubscription():
    def __init__(self,
                 exchange_name: str,
                 routing: str,
                 exchange_type: ExchangeType_=ExchangeType.TOPIC,
                 on_message: EVENT_CALLBACK_=None):
        self._routing = routing
        self._exchange_name = exchange_name
        self._exchange_type = exchange_type
        self._queue = None
        self.on_message = on_message

    def __str__(self):
        return f'<{self._routing} @ {self._exchange_name}>'

    @property
    def on_message(self) -> EVENT_CALLBACK_:
        return self._on_message

    @on_message.setter
    def on_message(self, f: EVENT_CALLBACK_):
        self._on_message = f if f else _default_on_message

    async def declare_on_remote(self, connection: Type[aio_pika.Connection]):
        LOGGER.info(f'Declaring event bus subscription {self} on {connection}')
        channel = await connection.channel()
        exchange = await channel.declare_exchange(self._exchange_name,
                                                  type=self._exchange_type,
                                                  auto_delete=True)
        self._queue = await channel.declare_queue(exclusive=True)
        await self._queue.bind(exchange, self._routing)
        await self._queue.consume(self._relay)

    async def _relay(self, message: Type[aio_pika.IncomingMessage]):
        """Relays incoming messages between the queue and the user callback"""
        message.ack()  # We always acknowledge, regardless of errors in client code
        try:
            await self.on_message(self, message.routing_key, json.loads(message.body))
        except Exception as ex:
            LOGGER.error(f'Exception relaying message in {self}: {ex}')


class EventListener():
    def __init__(self, app: Type[web.Application]=None):
        # Asyncio queues need a context loop
        # We'll initialize self._pending when we have one
        self._pending_pre_async = []
        self._pending = None
        self._connection = None
        self._task = None

        if app:
            self.setup(app)

    def __str__(self):
        return f'<{type(self).__name__} {self._connection}>'

    def setup(self, app):
        app.on_startup.append(self._startup)
        app.on_cleanup.append(self._cleanup)

    async def _startup(self, app: Type[web.Application]):
        await self.start(app.loop)

    async def _cleanup(self, app: Type[web.Application]):
        await self.close()

    def _on_connected(self, connection):
        if not connection.is_closed:
            LOGGER.info(f'Connected {self}')
            self._task = connection.loop.create_task(self._listen())

    async def start(self, loop: Type[asyncio.BaseEventLoop]):
        # Initialize the async queue now we know which loop we're using
        self._pending = asyncio.Queue(loop=loop)

        # Transfer all subscriptions that were made before the event loop started
        [self._pending.put_nowait(s) for s in self._pending_pre_async]

        # We won't be needing this anymore
        self._pending_pre_async = None

        # Create the connection
        # We want to start listening it now, and whenever we reconnect
        self._connection = await aio_pika.connect_robust(loop=loop, host=EVENTBUS_HOST)
        self._connection.add_reconnect_callback(self._on_connected)
        self._on_connected(self._connection)

    async def close(self):
        LOGGER.info(f'Closing {self}')

        if self._task:
            try:
                self._task.cancel()
                await self._task
            except CancelledError:
                pass

        if self._connection and not self._connection.is_closed:
            await self._connection.close()

    def subscribe(self,
                  exchange_name: str,
                  routing: str,
                  exchange_type: ExchangeType_=ExchangeType.TOPIC,
                  on_message: EVENT_CALLBACK_=None):
        """Adds a new event subscription to the listener.

        Actual queue declaration to the remote message server is done when connected.
        If the listener is not currently connected, it defers declaration.

        on_message(queue, key, message) can be specified.
        If the event was in JSON, message is a dict. Otherwise it is a string

        If on_message() is not set, it will default to logging the message.
        """
        sub = EventSubscription(
            exchange_name,
            routing,
            exchange_type,
            on_message=on_message
        )

        if self._pending is not None:
            self._pending.put_nowait(sub)
        else:
            self._pending_pre_async.append(sub)

        LOGGER.info(f'New event bus subscription: [{sub}]')
        return sub

    async def _listen(self):
        try:
            while not self._connection.is_closed:
                subscription = await self._pending.get()

                # Check whether connection wasn't closed while waiting for queue
                if not self._connection.is_closed:
                    await subscription.declare_on_remote(self._connection)

                else:  # pragma: no cover
                    # Connection closed while retrieving
                    # Put it back in the queue, we'll declare it after reconnect
                    self._pending.put_nowait(subscription)

        except RuntimeError as ex:  # pragma: no cover
            LOGGER.warn(f'Error in {self} {ex}')
        except CancelledError:
            LOGGER.info(f'Cancelled {self}')


class EventPublisher():
    def __init__(self, app: Type[web.Application]=None):
        self._connection = None
        self._channel = None

        if app:
            self.setup(app)

    def __str__(self):
        return f'<{type(self).__name__} {self._connection}>'

    def setup(self, app):
        app.on_startup.append(self._startup)
        app.on_cleanup.append(self._cleanup)

    async def _startup(self, app: Type[web.Application]):
        await self.start(app.loop)

    async def _cleanup(self, app: Type[web.Application]):
        await self.close()

    async def start(self, loop: Type[asyncio.BaseEventLoop]):
        def _on_connected(connection):
            if not connection.is_closed:
                LOGGER.info(f'Connected {self}')

        self._connection = await aio_pika.connect_robust(loop=loop, host=EVENTBUS_HOST)
        self._connection.add_reconnect_callback(_on_connected)
        _on_connected(self._connection)

    async def close(self):
        LOGGER.info(f'Closing {self}')
        if self._connection and not self._connection.is_closed:
            await self._connection.close()

    async def publish(self,
                      exchange: str,
                      routing: str,
                      message: Union[str, dict]='',
                      exchange_type: ExchangeType_=ExchangeType.TOPIC):

        # Makes for a more readable error in case of closed connections
        if not self._connection or self._connection.is_closed:
            raise ConnectionRefusedError(
                f'No event bus connection available for {self._connection}')

        # json.dumps() also correctly handles strings
        data = json.dumps(message).encode()

        if not self._channel:
            self._channel = await self._connection.channel()

        exchange = await self._channel.declare_exchange(
            exchange,
            exchange_type,
            auto_delete=True)

        # Push it over the line
        await exchange.publish(
            aio_pika.Message(data),
            routing_key=routing
        )


@routes.post('/_debug/publish')
async def post_publish(request):
    """
    ---
    tags:
    - Events
    summary: Publish event.
    description: Publish a new event message to the event bus.
    operationId: events.publish
    produces:
    - text/plain
    parameters:
    -
        in: body
        name: body
        description: Event message
        required: true
        schema:
            type: object
            properties:
                exchange:
                    type: string
                routing:
                    type: string
                message:
                    type: object
    """
    args = await request.json()
    try:
        await get_publisher(request.app).publish(
            args['exchange'],
            args['routing'],
            args['message']
        )
        return web.Response()

    except ConnectionRefusedError as ex:
        LOGGER.warn(f'Unable to publish {args}: {ex}')
        return web.Response(body='Event bus connection refused', status=500)



"""
Tests functionality offered by brewblox_service.events
"""

import json
from unittest.mock import Mock, call

import aio_pika
import pytest
from asynctest import Mock as AsyncMock
from asynctest import CoroutineMock

from brewblox_service import events

TESTED = events.__name__


def set_connect_func(mocker, closed: bool):

    async def mocked_connect_robust(loop=None, host=None):
        conn_mock = AsyncMock(spec=aio_pika.robust_connection.RobustConnection)
        conn_mock.loop = loop
        conn_mock.is_closed = closed

        chan_mock = CoroutineMock()
        exchange_mock = CoroutineMock()
        queue_mock = CoroutineMock()

        queue_mock.bind = CoroutineMock()
        queue_mock.consume = CoroutineMock()

        exchange_mock.publish = CoroutineMock()

        chan_mock.declare_exchange = CoroutineMock(return_value=exchange_mock)
        chan_mock.declare_queue = CoroutineMock(return_value=queue_mock)

        conn_mock.close = CoroutineMock()
        conn_mock.channel = CoroutineMock(return_value=chan_mock)

        return conn_mock

    mocker.patch.object(events.aio_pika, 'connect_robust',
                        mocked_connect_robust)


@pytest.fixture
async def app(app, mocker, loop):
    """App with events enabled"""
    set_connect_func(mocker, closed=False)

    events.setup(app)
    return app


async def test_setup(app, client):
    assert events.get_listener(app)
    assert events.get_publisher(app)


async def test_subscribe_callbacks(app, client, mocker):
    cb = Mock()
    content = dict(key='var')
    pika_msg = Mock()
    pika_msg.routing_key = 'message_key'
    pika_msg.body = json.dumps(content)

    sub = events.get_listener(app).subscribe(
        'brewblox', 'router', on_message=cb)

    await sub._relay(pika_msg)
    cb.assert_called_once_with(sub, 'message_key', content)
    assert pika_msg.ack.call_count == 1

    cb2 = Mock()
    sub.on_message = cb2

    await sub._relay(pika_msg)
    cb2.assert_called_once_with(sub, 'message_key', content)
    assert cb.call_count == 1
    assert pika_msg.ack.call_count == 2

    # Shouldn't blow up
    cb2.side_effect = ValueError
    await sub._relay(pika_msg)

    # Should be tolerated
    sub.on_message = None
    await sub._relay(pika_msg)


async def test_offline_listener(app, mocker):
    # We're not importing the client fixture, so the app is not running
    # Expected behaviour is for the listener to add functions to the app hooks
    listener = events.EventListener(app)

    assert listener._startup in app.on_startup
    assert listener._cleanup in app.on_cleanup

    # Subscriptions will not be declared yet - they're waiting for startup
    sub = listener.subscribe('exchange', 'routing')
    assert sub in listener._pending_pre_async

    # Can safely be called, but will be a no-op at this time
    await listener.close()


async def test_online_listener(app, client, mocker):
    # The client fixture has started running the app in the event loop
    # We'll need to start explicitly
    with pytest.raises(RuntimeError):
        events.EventListener(app)

    listener = events.EventListener()
    sub = listener.subscribe('exchange', 'routing')

    # No-op, listener is not yet started
    await listener.close()

    assert sub in listener._pending_pre_async
    await listener.start(app.loop)
    assert listener._pending_pre_async is None

    pending_subs = listener._pending.qsize()
    listener.subscribe('exchange', 'routing')
    # We haven't yielded, so the subscription can't have been processed yet
    assert listener._pending.qsize() == pending_subs + 1

    # Safe for repeated calls
    await listener.close()
    await listener.close()


async def test_closed_listener(app, client, mocker):
    set_connect_func(mocker, closed=True)

    listener = events.EventListener()
    await listener.start(app.loop)


async def test_disconnect_listener(app, client, mocker):
    listener = events.EventListener()
    await listener.start(app.loop)

    # Task should be started
    assert listener._task

    # Close connection. Task should gracefully exit now
    listener._connection.is_closed = True
    await listener._task


async def test_offline_publisher(app):
    publisher = events.EventPublisher(app)

    assert publisher._startup in app.on_startup
    assert publisher._cleanup in app.on_cleanup

    with pytest.raises(ConnectionRefusedError):
        await publisher.publish('exchange', 'key', message=dict(key='val'))


async def test_online_publisher(app, client, mocker):
    # Can't setup while running
    with pytest.raises(RuntimeError):
        events.EventPublisher(app)

    publisher = events.EventPublisher()
    await publisher.start(app.loop)

    await publisher.publish('exchange', 'key', message=dict(key='val'))
    await publisher.publish('exchange', 'key', message=dict(key='val'))


async def test_closed_publisher(app, client, mocker):
    set_connect_func(mocker, closed=True)

    publisher = events.EventPublisher()
    await publisher.start(app.loop)


async def test_publish_endpoint(app, client, mocker):
    publish_spy = mocker.spy(events.get_publisher(app), 'publish')

    # standard ok
    assert (await client.post('/_debug/publish', json=dict(
        exchange='exchange',
        routing='first',
        message=dict(key='val')
    ))).status == 200

    # string messages supported
    assert (await client.post('/_debug/publish', json=dict(
        exchange='exchange',
        routing='second',
        message='message'
    ))).status == 200

    # return 500 on connection refused
    events.get_publisher(app)._connection.is_closed = True
    assert (await client.post('/_debug/publish', json=dict(
        exchange='exchange',
        routing='third',
        message='message'
    ))).status == 500

    publish_spy.assert_has_calls([
        call('exchange', 'first', dict(key='val')),
        call('exchange', 'second', 'message'),
        call('exchange', 'third', 'message'),
    ])
